/*
  intelhex.c
  Matlib 2010-10-19
*/

#include "intelhex.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>


static int ih_new (MemoryBlock *blk, uint32_t start, uint32_t length)
{
  if (*blk != NULL)
    ih_free (blk);
  *blk = malloc (sizeof (struct memory_block) + length);
  if (*blk == NULL)
  {
    printf ("intelhex: cannot allocate %u bytes of memory\n", 
       sizeof (struct memory_block) + length);
    return -1;
  }
  (*blk) -> start = start;
  (*blk) -> length = length;
  (*blk) -> next = NULL;
  return 0;
}

static void ih_free (MemoryBlock *blks)
{
  MemoryBlock pb, cb;
  pb = *blks;
  while (pb != NULL)
  {
    cb = pb;
    pb = pb -> next;
    free (cb);
  }
  *blks = NULL;
}



#define IH_DATA 0
#define IH_EOF  1
#define IH_ESA  2
#define IH_SSA  3
#define IH_ELA  4
#define IH_SLA  5



#define IHERRORIF(__COND__,__TEXT__) \
  if (__COND__) \
  { \
    printf ("intelhex: %s (%s L%u C%u)\n", __TEXT__, name, line, col); \
    fclose (hf); \
    return -1; \
  }
  
#define IHLOADCHAR() \
  { \
    col ++; \
    IHERRORIF ((c = fgetc (hf)) == EOF, "unexpected end of file"); \
  }
  
#define IHLOADBYTE() \
  IHLOADCHAR (); \
  if (c >= 'a') \
    c -= 32; \
  IHERRORIF (((c < '0') || (c > '9')) && ((c < 'A') || (c > 'F')), \
     "hex code expected"); \
  if (c < 'A') \
    b = c - '0'; \
  else \
    b = c - 'A' + 10; \
  b <<= 4; \
  IHLOADCHAR (); \
  if (c >= 'a') \
    c -= 32; \
  IHERRORIF (((c < '0') || (c > '9')) && ((c < 'A') || (c > 'F')), \
     "hex code expected"); \
  if (c < 'A') \
    b += c - '0'; \
  else \
    b += c - 'A' + 10; \
  cs += b

static int ih_loadfile (char *name, MemoryBlock *blks)
{
  FILE *hf;
  int c;
  unsigned line, col;
  unsigned char b, cs, btype;
  uint32_t ihl, iha, address;
  MemoryBlock *cb;
  
  if (*blks != NULL)
    ih_free (blks);
  line = 1;
  if ((hf = fopen (name, "r")) == NULL)
  {
    printf ("intelhex: cannot open file %s: %s\n", name, strerror (errno));
    return -1;
  }
  address = 0;
  do
  {
    col = 0;
    IHLOADCHAR ();
    IHERRORIF (c != ':', "':' expected");
    cs = 0;
    IHLOADBYTE ();
    ihl = b;
    IHLOADBYTE ();
    iha = b << 8;
    IHLOADBYTE ();
    iha |= b;
    address += iha;
    IHLOADBYTE ();
    btype = b;
    switch (btype)
    {
      case IH_DATA:
        for (cb = blks; *cb != NULL; cb = &((*cb) -> next))
          if ((*cb) -> start + (*cb) -> length == address)
            break;
        if (*cb == NULL)
        {
          *cb = malloc (sizeof (struct memory_block) +  ihl);
          (*cb) -> start = address;
          (*cb) -> length = 0;
          (*cb) -> next = NULL;
        }
        else
        {
          ihl += (*cb) -> length;
          *cb = realloc (*cb, sizeof (struct memory_block) + 
             ihl);
        }
        if (*cb == NULL)
        {
          printf ("intelhex: cannot allocate %u bytes of memory (%s L%u C%u)\n",
            sizeof (struct memory_block) + ihl + (*cb) -> length, 
            name, line, col);
          fclose (hf);
          return -1;
        }
        while ((*cb) -> length < ihl)
        {
          IHLOADBYTE ();
          (*cb) -> data [(*cb) -> length ++] = b;
        }
        address = 0;
        break;
//    case IH_EOF: // EOF
      case IH_ESA: // 20 bit address segment
        IHERRORIF (ihl != 2, "length not 2 in ESA record");
        IHERRORIF (iha != 0, "address not 0 in ESA record");
        IHLOADBYTE ();
        address = b << 12;
        IHLOADBYTE ();
        address |= b  << 4;
        break;
      case IH_SSA: // 20 bit full address
        IHERRORIF (ihl != 0, "length not 4 in SSA record");
        IHERRORIF (iha != 0, "address not 0 in SSA record");
        IHLOADBYTE ();
        address = b << 12;
        IHLOADBYTE ();
        address |= b  << 4;
        IHLOADBYTE ();
        address += b << 8;
        IHLOADBYTE ();
        address += b;
        break;
      case IH_ELA: //  32 bit address
        IHERRORIF (ihl != 2, "length not 2 in ELA record");
        IHERRORIF (iha != 0, "address not 0 in ELA record");
        IHLOADBYTE ();
        address = b << 24;
        IHLOADBYTE ();
        address |= b << 16;
        break;
      case IH_SLA: // 32 bit full address
        IHERRORIF (ihl != 0, "length not 4 in SLA record");
        IHERRORIF (iha != 0, "address not 0 in SLA record");
        IHLOADBYTE ();
        address = b << 24;
        IHLOADBYTE ();
        address |= b  << 16;
        IHLOADBYTE ();
        address |= b << 8;
        IHLOADBYTE ();
        address |= b;
        break;
    }
    IHLOADBYTE ();
    IHERRORIF (cs != 0, "incorrect checksum");
    IHLOADCHAR ();
    if (c == '\r')
      IHLOADCHAR ();
    IHERRORIF (c != '\n', "new line expected");
    line ++;
  }
  while (btype != 1);
  fclose (hf);
  return 0;
}

#undef IHERRORIF
#undef IHLOADCHAR
#undef IHLOADBYTE



static int ih_writerecord (FILE *f, uint8_t length, uint16_t address,
   unsigned char type, unsigned char *data)
{
  unsigned char cs;
  
  ERROR_IF (fprintf (f, ":%02X%04X%02X", length, address, type) < 0,
     "cannot write to file: %s", strerror (errno));
  cs = length + (address >> 8) + (address & 0xff) + type;
  while (length > 0)
  {
    ERROR_IF (fprintf (f, "%02X", *data) < 0,
       "cannot write to file: %s", strerror (errno));
    cs += *data;
    length --;
    data ++;
  }
  ERROR_IF (fprintf (f, "%02X\r\n", (unsigned char) (- cs)) < 0,
     "cannot write to file: %s", strerror (errno));
  return 0;
}


static int ih_savefile (char *name, MemoryBlock blks)
{
  FILE *f;
  uint32_t p, s;
  
  ERROR_IF ((f = fopen (name, "w")) == NULL,
     "Cannot open file '%s' for writing: %s", name, strerror (errno));
  while (blks != NULL)
  {
    if (((blks -> start & 0xffff0000) != 0) && 
       (ih_writerecord (f, 2, 0, IH_ELA, (void *) &(blks -> start)) != 0))
      return -1;
    for (p = 0; p < blks -> length; p += s)
    {
      s = blks -> length - p;
      if (s > 128)
        s = 128;
      if (ih_writerecord (f, s, blks -> start + p, IH_DATA,
         blks -> data + p) != 0)
        return -1;
    }
    blks = blks -> next;
  }
  if (ih_writerecord (f, 0, 0, IH_EOF, NULL) != 0)
    return -1;
  fclose (f);
  return 0;
}


