/*
  tiny45910.c
  Matlib 2010-10-15
*/

#include "tpi.h"
#include "intelhex.h"

typedef struct
{
  char *name;
  unsigned char id1, id2;
  uint16_t lock_bits, conf_bits, calibration_bits, code;
  unsigned rom_size, ram_size;
} DeviceType;



#define DEVICES 4
static const DeviceType devices [DEVICES] =
{
  {"ATtiny4",  0x8f, 0x0a, 0x3f00, 0x3f40, 0x3f80, 0x4000, 512,  96},
  {"ATtiny5",  0x8f, 0x09, 0x3f00, 0x3f40, 0x3f80, 0x4000, 512,  96},
  {"ATtiny9",  0x90, 0x08, 0x3f00, 0x3f40, 0x3f80, 0x4000, 1024, 96},
  {"ATtiny10", 0x90, 0x03, 0x3f00, 0x3f40, 0x3f80, 0x4000, 1024, 96}
};

const DeviceType *device;


#define MANUFS 1
static const NameID manufs [MANUFS] =
{
  {"Atmel", NULL, 0x1e}
};



static int tiny_deviceid ()
{
  unsigned char manuf, id1, id2;
  unsigned i;

  printf ("Checking device... ");
  tpi_sstpr (0x3fc0);
  if ((tpi_sldp (&manuf) == 0) && (tpi_sldp (&id1) == 0)
     && (tpi_sld (&id2) == 0))
  {
    printf ("done\n  Manufacturer: 0x%02x   ", manuf);
    i = 0;
mloop:
    if (i == MANUFS)
      puts ("unknown");
    else if (manufs [i].id == manuf)
      puts (manufs [i].name);
    else
    {
      i ++;
      goto mloop;
    }
    
    printf (      "  Device:       0x%02x%02x ", id1, id2);
    i = 0;
    for (device = devices; (device -> id1 != id1) || (device -> id2 != id2);
       device ++)
      if (++ i == DEVICES)
      {
        puts ("unknown");
        return -1;
      }
    printf ("%s (flash %u bytes)\n", device -> name, device -> rom_size);
  }
  return 0;
}




// flash writing commands

#define NO_OPERATION   0x00
#define CHIP_ERASE     0x10
#define SECTION_ERASE  0x14
#define WORD_WRITE     0x1d

#define NVMBSY  0x80

static int tiny_setwrite (unsigned char mode)
{
  return tpi_sout (0x33, mode);
}

static int tiny_waitnvm (unsigned interval)
{
  unsigned char nvmcsr;
loop:
  if (tpi_sin (0x32, &nvmcsr) != 0)
    return -1;
  if ((nvmcsr & NVMBSY) != 0)
  {
    msleep (interval);
    goto loop;
  }
  return 0;
}

static int tiny_chiperase ()
{
  printf ("Erasing flash memory... ");
  if (tiny_setwrite (CHIP_ERASE) != 0)
    return -1;
  tpi_sstpr (device -> code | 1);
  tpi_sstp (0xff);
  if ((tiny_waitnvm (500) != 0) || (tiny_setwrite (NO_OPERATION) != 0))
    return -1;
  printf ("done\n");
  return 0;
}

static int tiny_sectionerase (uint16_t block)
{
  printf ("Erasing section @ 0x%04X... ", block);
  if (tiny_setwrite (SECTION_ERASE) != 0)
    return -1;
  tpi_sstpr (block | 1);
  tpi_sstp (0xff);
  if ((tiny_waitnvm (500) != 0) || (tiny_setwrite (NO_OPERATION) != 0))
    return -1;
  printf ("done\n");
  return 0;  
}





static const NameID tiny_configbits [8] =
{
  {"RESET disabled",     "RSTDISBL", 0x01},
  {"Watchdog always on", "WDTON",    0x02},
  {"Clock output",       "CKOUT",    0x04},
  {"Unknown/invalid",    "",         0},
};

static int tiny_readconfig (unsigned char *cb)
{
  printf ("Reading config bits... ");
  if ((tpi_sstpr (device -> conf_bits) != 0) || (tpi_sld (cb) != 0))
    return -1;
  printf ("0x%02X\n", *cb);
  return 0;
}

static int tiny_writeconfig (const unsigned char cb)
{
  unsigned char old_cb;
  
  if (tiny_readconfig (&old_cb) != 0)
    return -1;
  if ((cb != (old_cb & cb)) && (tiny_sectionerase (device -> conf_bits) != 0))
    return -1;
  printf ("Writing config bits... ");
  if ((tiny_setwrite (WORD_WRITE) != 0) 
     || (tpi_sstpr (device -> conf_bits) != 0) 
     || (tpi_sstp (cb) != 0) || (tpi_sstp (0xff) != 0) 
     || (tiny_waitnvm (10) != 0))
    return -1;
  printf ("0x%02X\n", cb);
  return 0;
}




static const NameID tiny_lockbits [8] =
{
  {"Flash write disabled", "NVLB1", 0x01},
  {"Flash read disabled",  "NVLB2", 0x02},
  {"Unknown/invalid",      "",         0},
};

static int tiny_readlock (unsigned char *lb)
{
  printf ("Reading lock bits... ");
  if ((tpi_sstpr (device -> lock_bits) != 0) || (tpi_sld (lb) != 0))
    return -1;
  printf ("0x%02X\n", *lb);
  return 0;
}

static int tiny_writelock (unsigned char lb)
{
  printf ("Writing lock bits... ");
  if ((tiny_setwrite (WORD_WRITE) != 0) 
     || (tpi_sstpr (device -> lock_bits) != 0) 
     || (tpi_sstp (lb) != 0) || (tpi_sstp (0xff) != 0)
     || (tiny_waitnvm (10) != 0))
    return -1;
  printf ("0x%02X\n", lb);
  return 0;
}




static int tiny_readmem (MemoryBlock *mb, uint16_t start, uint16_t length)
{
  uint32_t p;
  
  stat_reset ();
  printf ("Reading memory data @ 0x%04X...\n", start);
  if (ih_new (mb, 0, length) != 0)
    return -1;
  tpi_sstpr (start);
  p = 0;
  while (p < length)
  {
    if (tpi_sldp ((*mb) -> data + p) != 0)
      return -1;
    p ++;
    ARG_STAT (start + p, p, length);
  }
  return 0;
}

static int tiny_readram (MemoryBlock *mb)
{
  return tiny_readmem (mb, 0, device -> ram_size);
}

static int tiny_readcode (MemoryBlock *mb)
{
  return tiny_readmem (mb, device -> code, device -> rom_size);
}

static int tiny_erasecode ()
{
  return tiny_sectionerase (device -> code);
}

static int tiny_writecode (MemoryBlock blks)
{
  MemoryBlock cb;
  unsigned p, pr;
  unsigned char d [2], r [2];
  
  stat_reset ();
  printf ("Checking data to be written... ");
  p = 0;
  for (cb = blks; cb != NULL; cb = cb -> next)
  {
    p ++;
    if (cb -> start > device -> rom_size)
    {
      printf ("0x%08X: start address out of range\n", cb -> start);
      return -1;
    }
    else if (cb -> start + cb -> length > device -> rom_size)
    {
      printf ("0x%08X: data beyond flash size\n", cb -> start);
      return -1;
    }
  }
  printf ("%u blocks\n", p);
  if (p == 0)
    return 0;
  if (tiny_sectionerase (device -> code) != 0)
    return -1;
  printf ("Writing data...\n");
  if (tiny_setwrite (WORD_WRITE) != 0)
    return -1;
  for (cb = blks; cb != NULL; cb = cb -> next)
  {
    printf ("  0x%04X size %u:\n", cb -> start, cb -> length);
    p = 0;
    pr = cb -> start + device -> code;
    if ((pr & 1) == 0)
      tpi_sstpr (pr);
    else
    {
      tpi_sstpr (-- pr);
      if (tpi_sldp (d) != 0)
        return -1;
      goto second;
    }
first:
    if (p == cb -> length)
      continue;
    d [0] = cb -> data [p ++];
second:
    if (p == cb -> length)
    {
      tpi_sstpr (pr + 1);
      if (tpi_sld (d + 1) != 0)
        return -1;
      tpi_sstpr (pr);
    }
    else
      d [1] = cb -> data [p ++];
    tpi_sstp (d [0]);
    tpi_sst (d [1]);
    if (tiny_waitnvm (10) != 0)
      return -1;
    tpi_sstpr (pr);
    if ((tpi_sldp (r) != 0) || (tpi_sldp (r + 1) != 0))
      return -1;
    if ((d [0] != r [0]) || (d [1] != r [1]))
    {
      printf ("verification failed: data=0x%02X%02X read=0x%02X%02X\n",
         d [1], d [0], r [1], r [0]);
      return -1;
    }
    pr += 2;
    ARG_STAT (pr, p, cb -> length);
    goto first;
  }
  if (tiny_setwrite (NO_OPERATION) != 0)
    return -1;
  printf ("done\n");
  return 0;
}

