/*
  tpiprog.c
  Matlib 2010-10-09

   THIS SOFTWARE IS IN PUBLIC DOMAIN
     There is no limitation on any kind of usage of this code. If you
     modify this code please remove this header and rename the file.

     You use it because it is yours.

     Originally written and maintained by Matlib (matlib@laser.t28.net),
     any comments or suggestions are welcome.

   Change log:
        1  2010-12-15  Matlib
*/

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>


#include <arguments.h>

#include "../config.h"

struct Argument arguments [];

enum
{
#ifdef PPDIRECT
  ARG_N_DPRT,
#endif
#ifdef PPDEV
  ARG_N_PDEV,
#endif

  ARG_N_CLBR, 
  ARG_N_USLN,
  ARG_N_CLKT,
  ARG_N_TPIE,
  ARG_N_STAT,
  ARG_N_SFLW,
  ARG_N_IERR,
  ARG_N_IRAC
};

#define ARG_DPRT ((char *) (arguments [ARG_N_DPRT].value))
#define ARG_USLN ((unsigned long) (arguments [ARG_N_USLN].value))
#define ARG_CLKT ((unsigned long) (arguments [ARG_N_CLKT].value))
#define ARG_TPIE ((void (*) ()) (arguments [ARG_N_TPIE].value))
#define ARG_STAT ((void (*) (unsigned, unsigned, unsigned)) (arguments [ARG_N_STAT].value))
#define ARG_IERR (arguments [ARG_N_IERR].value != NULL)
#define ARG_IRAC (arguments [ARG_N_IRAC].value != NULL)



#include "util.c"
#include "intelhex.c"



static int (*hw_setport) (unsigned, unsigned);
static int (*hw_readport) (unsigned, int *);
static int (*hw_power) (int);
static int (*hw_reset) (int);
static unsigned hw_ports = 0;

static void check_driver ()
{
  if (hw_ports != 0)
    arguments_error_freeform ("Only one driver may be used.");
}



#include "tpi.c"

#ifdef PPDIRECT
  #include "ppdirect.c"
#endif
#ifdef PPDEV
  #include "ppdev.c"
#endif

#include "tiny45910.c"



// turn device off if programming interrupted

static void turn_off ()
{
  unsigned i;
  printf ("Shutting device down... ");
  for (i = 0; i < 4; i ++)
    if (hw_setport (i, HW_HIZ) != 0)
    {
      puts ("\nError while switching outputs to high-impedance mode.");
      return;
    }
  if ((hw_reset (0) != 0) || (hw_power (0) != 0))
  {
    puts ("\nError while turning power off.");
    return;
  }
  printf ("done\n");
}

void sigterm (int signum)
{
  printf ("\n\nInterrupted.\n");
  exit (EXIT_FAILURE);
}



struct Command
{
  const char *name;
  const int (*proc) (char *);
  const char *help;
};

static const struct Command commands [];
MemoryBlock main_mb;


static const int show_help (char *arg)
{
  unsigned i, j, l;
  if (*arg == 0)
  {
    printf ("\nCommand syntax:\n  command[=argument]\n\nCommand list:\n");
    l = 0;
    for (i = 0; commands [i].name != NULL; i ++)
    {
      j = strlen (commands [i].name);
      if (j > l)
        l = j;
    }
    for (i = 0; commands [i].name != NULL; i ++)
    {
      printf ("  ");
      printf (commands [i].name);
      for (j = l - strlen (commands [i].name); j > 0; j --)
        putchar (' ');
      printf ("  ");
      puts (commands [i].help);
    }
  }
  return 0;
}

static const int show_configbits (char *arg)
{
  show_bits (tiny_configbits);
  return 0;
}

static const int read_configbits (char *arg)
{
  unsigned char cb;
  if (tiny_readconfig (&cb) != 0)
    return -1;
  show_bits_set (tiny_configbits, cb, "none");
  return 0;
}

static const int set_configbits (char *arg)
{
  unsigned char cb;
  if ((set_bits (tiny_configbits, arg, &cb) != 0) 
     || (tiny_writeconfig (cb) != 0))
    return -1;
  return 0;
}


static const int show_lockbits (char *arg)
{
  show_bits (tiny_lockbits);
  return 0;
}

static const int read_lockbits (char *arg)
{
  unsigned char lb;
  if (tiny_readlock (&lb) != 0)
    return -1;
  show_bits_set (tiny_lockbits, lb, "no locks");
  return 0;
}

static const int set_lockbits (char *arg)
{
  unsigned char lb;  
  if ((set_bits (tiny_lockbits, arg, &lb) != 0) 
     || (tiny_writelock (lb) != 0))
    return -1;
  return 0;
}

static const int loadfile (char *arg)
{
  printf ("Loading file '%s'... ", arg);
  if (ih_loadfile (arg, &main_mb) != 0)
    return -1;
  printf ("done\n");
  return 0;
}

static const int writefile (char *arg)
{
  printf ("Writing to file '%s'... ", arg);
  if (ih_savefile (arg, main_mb) != 0)
    return -1;
  printf ("done\n");
  return 0;
}

static const int dev_send (char *arg)
{
  return tiny_writecode (main_mb);
}

static const int dev_get (char *arg)
{
  long start, length;
  char *tail;
  
  if (*arg == 0)
    return tiny_readcode (&main_mb);
  if (strcasecmp (arg, "ram") == 0)
    return tiny_readram (&main_mb);
  start = strtol (arg, &tail, 0);
  ERROR_IF (*tail == 0, "Length is missing.");
  ERROR_IF (*tail != ',', "Invalid number format for start address: '%s'.", 
     arg);
  tail ++;
  length = strtol (tail, &tail, 0);
  ERROR_IF (*tail != 0, "Invalid number format for length: '%s'.", arg);
  ERROR_IF ((start < 0) || (start > 0xffff), 
     "Start address 0x%lX out of range.", start);
  ERROR_IF ((length < 0) || (length + start > 0xffff), 
     "Length 0x%04lX + %lu out of range.", start, length);
  return tiny_readmem (&main_mb, start, length);
}

static const int dev_chiperase (char *arg)
{
  return tiny_chiperase ();
}

static const int dev_sectionerase (char *arg)
{
  unsigned long start;
  char *tail;
  
  if (*arg == 0)
    return tiny_erasecode ();
  start = strtol (arg, &tail, 0);
  ERROR_IF (*tail != 0, "Invalid number format for section address: '%s'.", 
     arg);
  ERROR_IF ((start < 0) || (start > 0xffff), 
     "Section address 0x%lX out of range", start);
  return tiny_sectionerase (start);
}

static const int show_mb (char *arg)
{
  unsigned char l;
  unsigned i, e;
  MemoryBlock mb;
  for (mb = main_mb; mb != NULL; mb = mb -> next)
  {
    printf ("\nStart: 0x%08X Length: %u\n ", mb -> start, mb -> length);
    if (mb -> length > 0)
    {
      l = mb -> data [mb -> length - 1];
      e = mb -> length - 1;
      while (e > 0)
        if (mb -> data [-- e] != l)
        {
          e ++;
          break;
        }
      for (i = 0; i <= e; i ++)
        printf (" %02X", mb -> data [i]);
      if (i != mb -> length)
        printf (" [%02X]", l);
    }
    printf ("\n"); 
  }
  return 0;
}



static const int interactive (char *arg)
{
  static const char K_ON [10]  = "1234567890";
  static const char K_OFF [10] = "qwertyuiop";
  static const char K_Z [10]   = "asdfghjkl;";
  char c;
  unsigned i;
  long flags;
  int value;
  struct termios attr, attr_s;

  printf ("Entering interactive mode\n");
  if (isatty (STDIN_FILENO))
  {
    SYS_ERROR (tcgetattr (STDIN_FILENO, &attr_s) != 0, "tcgetattr",
       "Cannot get terminal attributes.");
    attr = attr_s;
    attr.c_lflag &= ~(ICANON | ECHO);
    attr.c_cc [VMIN] = 1;
    attr.c_cc [VTIME] = 0;
    SYS_ERROR (tcsetattr (STDIN_FILENO, TCSAFLUSH, &attr) != 0, "tcsetattr",
       "Cannot set terminal attributes.");
  }
  SYS_ERROR ((flags = fcntl (STDIN_FILENO, F_GETFL, 0)) == -1, "fcntl", 
     "Cannot read input stream status flags.");
  SYS_ERROR (fcntl (STDIN_FILENO, F_SETFL, flags | O_NONBLOCK) != 0, "fcntl",
     "Cannot set input stream status flags.");

loop:
  value = -1;
  msleep (10);
  for (i = 0; i < hw_ports; i ++)
  {
    if (hw_readport (i, &value) != 0)
      goto end;
    if (value == 0)
      printf ("[  ] ");
    else
      printf ("[P%u] ", i + 1);
  }
  putchar ('\r');
  value = 0;

  if (read (STDIN_FILENO, &c, 1) == 1)
  {
    switch (c)
    {
      case '-': hw_power (0); break;
      case '=': hw_power (1); break;
      case '_': hw_reset (0); break;
      case '+': hw_reset (1); break;
      case '?':
        printf ("- =         power off and on\n"
                "_ +         +12 V off and on\n");
        for (i = 0; i < 10; i ++)
          if (i < hw_ports)
            putchar (K_ON [i]);
          else
            putchar (' ');
        printf (          "  force high level\n");
        for (i = 0; i < 10; i ++)
          if (i < hw_ports)
            putchar (K_OFF [i]);
          else
            putchar (' ');
        printf (          "  force low level\n");
        for (i = 0; i < 10; i ++)
          if (i < hw_ports)
            putchar (K_Z [i]);
          else
            putchar (' ');
        printf (          "  high impedance\n"
                "x           exit\n\n");
        break;
      default:
        for (i = 0; i < hw_ports; i ++)
          if (c == K_ON [i])
            value = hw_setport (i, HW_ON);
          else if (c == K_OFF [i])
            value = hw_setport (i, HW_OFF);
          else if (c == K_Z [i])
            value = hw_setport (i, HW_HIZ);
    }
    if ((c != 'x') && (value == 0))
      goto loop;
  }
  else if (errno == EWOULDBLOCK)
    goto loop;

end:
  if (isatty (STDIN_FILENO))
    SYS_ERROR (tcsetattr (STDIN_FILENO, TCSAFLUSH, &attr_s) != 0, "tcsetattr",
       "Cannot set terminal attributes.");
  SYS_ERROR (fcntl (STDIN_FILENO, F_SETFL, flags) != 0, "fcntl",
     "Cannot set input stream status flags.");
  printf ("\nLeaving interactive mode\n");
  if (value == 0)
    value = tpi_init ();
  return value;
}




static const struct Command commands [] =
{
  {"?",    show_help,         "show help screen"},
  {"cb",   read_configbits,   "read config bits"},
  {"scb",  set_configbits,    "set config bits; scb, scb=<value>, scb=<name>,..."},
  {"vcb",  show_configbits,   "show available config bits"},
  {"lb",   read_lockbits,     "read lock bits"},
  {"slb",  set_lockbits,      "set lock bits; slb=<value>, slb=<name>,..."},
  {"vlb",  show_lockbits,     "show available lock bits"},
  {"view", show_mb,           "view loaded code"},
  {"lf",   loadfile,          "load data from file; lf=<name>"},
  {"wf",   writefile,         "write data to file; wf=<name>"},
  {"get",  dev_get,           "load data from device; get, get=ram, get=<address>,<length>"},
  {"send", dev_send,          "write code to device"},
  {"se",   dev_sectionerase,  "erase section; se, se=<address>"},
  {"ce",   dev_chiperase,     "erase chip"},
  {"int",  interactive,       "enter interactive mode"},
  {}
};

int cmd_exec (char *cmd)
{
  unsigned i, p;
  
  if (*cmd == 0)
    return 0;
  for (p = 0; (cmd [p] != 0) && (cmd [p] != '='); p ++);
  
  i = 0;
  while ((strlen (commands [i].name) != p) 
     || (strncasecmp (cmd, commands [i].name, p) != 0))
    if (commands [++ i].name == NULL)
    {
      printf ("Invalid command '%s'\n", cmd);
      return ARG_IERR ? 0 : -1;
    }
  if (cmd [p] == '=')
    p ++;
  if (ARG_IERR)
  {
    commands [i].proc (cmd + p);
    return 0;
  }
  return commands [i].proc (cmd + p);
}

int cmd_from_file (FILE *f)
{
  char *buffer;
  size_t bs, bp;
  int c;
  
  bs = 1024;
  ERROR_IF ((buffer = malloc (bs)) == NULL, 
    "Cannot allocate %u bytes of memory: %s\n", bs, strerror (errno));
  
bloop:
  bp = 0;
loop:
  c = fgetc (f);
  if ((c == EOF) || (c == ' ') || (c == '\n'))
  {
    buffer [bp] = 0;
    ERROR_IF (cmd_exec (buffer) != 0, "Script interrupted\n");
    if (c == EOF)
      return 0;
    goto bloop;
  }
  buffer [bp ++] = c;
  if (bp == bs)
  {
    bs += 1024;
    ERROR_IF ((buffer = realloc (buffer, bs)) == NULL,
       "Cannot allocate %u bytes of memory: %s\n", bs, strerror (errno));
  }
  goto loop;
}



struct Argument arguments [] =
{
#ifdef PPDIRECT
  {'p', "direct-port",   ppdirect,         ARGUMENT_F_ONCE | ARGUMENT_F_PAR, NULL},
#endif
#ifdef PPDEV
  {'d', "ppdev",         ppdev,            ARGUMENT_F_ONCE | ARGUMENT_F_PAR, NULL},
#endif
  {'c', "calib-us",      usleep_calibrate, ARGUMENT_F_ONCE, NULL},
  {'n', "calib-n",       usleep_setn,      ARGUMENT_F_ONCE | ARGUMENT_F_PAR, NULL},
  {'T', "clock-t",       argument_clkt,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) 50},
  {'H', "hi-voltage",    tpi_hiv,          ARGUMENT_F_ONCE, tpi_enter_lowv},
  {'s', "stat-progress", status_progress,  ARGUMENT_F_ONCE, NULL},
  {'S', "stat-flow",     status_flow,      ARGUMENT_F_ONCE, NULL},
  {'e', "ignore-errors", argument_bool,    ARGUMENT_F_ONCE, NULL},
  {'i', "interactive",   argument_bool,    ARGUMENT_F_ONCE, NULL},
  {}
};

int main (int argc, char **argv)
{
  int argp;
  
  arguments [ARG_N_STAT].value = isatty (STDIN_FILENO) ? status_bar_progress 
     : status_bar_flow;
  argp = arguments_parse ("tpiprog", "1 (2010-12-15)", "[command]...", 
     "Parallel port TPI flash programmer", argc, argv, arguments);
  
  if (hw_ports == 0) 
#ifdef DEFAULT_PPDEV
    ppdev (argc, argv, NULL, "/dev/parport0");
#else
    ppdirect (argc, argv, NULL, "0x378");
#endif

  setbuf (stdout, NULL);
  signal (SIGINT, sigterm);
  signal (SIGQUIT, sigterm);
  signal (SIGTERM, sigterm);
  atexit (turn_off);

  if ((ARG_IRAC && (interactive (NULL) != 0)) 
     || (!ARG_IRAC && (tpi_init () != 0))
     || (tiny_deviceid () != 0))
    return EXIT_FAILURE;
  putchar ('\n');
  if (argp != argc)
    while (argp < argc) 
    {
      if (cmd_exec (argv [argp ++]) != 0)
        return EXIT_FAILURE;
    }
  else
  {
    if (isatty (STDIN_FILENO))
      arguments [ARG_N_IERR].value = (void *) -1;
    if (cmd_from_file (stdin) != 0)
      return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}

