/*
  util.c
  Matlib 2010-11-12
*/

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#define ERROR_IF(__COND__,...) \
  if (__COND__) \
  { \
    printf (__VA_ARGS__); \
    putchar ('\n'); \
    return -1; \
  }
  
#define SYS_ERROR(__COND__,__FUNC__,...) \
  if (__COND__) \
  { \
    printf ("%s: %s: %s\n", __func__, __FUNC__, strerror (errno)); \
    printf (__VA_ARGS__); \
    putchar ('\n'); \
    exit (EXIT_FAILURE); \
  }
  
#define SYS_BREAK(__COND__,__FUNC__)  ERROR_IF (__COND__, "%s: %s: %s", \
   __func__, __FUNC__, strerror (errno))




typedef struct
{
  const char *name;
  const char *short_name;
  const unsigned char id;
} NameID;




#include <unistd.h>

void msleep (unsigned t)
{
  usleep (t * 1000);
}



int (*usleep_cal) (useconds_t) = usleep;



#include <sys/timex.h>

int usleep_loop (useconds_t t)
{
  volatile unsigned long i; // prevent loop from being optimized
  for (i = t * ARG_USLN; i > 0; i --);
  return 0;
}

ARGUMENT_CALLBACK (usleep_calibrate)
{
  volatile unsigned long i;
  unsigned long n;
  struct ntptimeval tv1, tv2;
  
  printf ("Calibrating µ-second loop...");
  if (ntp_gettime (&tv1) != 0)
    printf ("\n  ntp_gettime returns error, ignoring");
  n = 100000;
loop:
  ntp_gettime (&tv1);
  for (i = n; i > 0; i --);
  ntp_gettime (&tv2);
  if (tv1.time.tv_usec > tv2.time.tv_usec)
  {
    i = 1000000 - tv1.time.tv_usec + tv2.time.tv_usec;
    i += (tv2.time.tv_sec - tv1.time.tv_sec - 1) * 1000000;
  }
  else
  {
    i = tv2.time.tv_usec - tv1.time.tv_usec;
    i += (tv2.time.tv_sec - tv1.time.tv_sec) * 1000000;
  }
  printf ("\n  %lu µs/%lu clks...", i, n);
  if (i < 1000000)
  {
    i = n;
    n *= 10;
    if (n < i) // we've just exceeded range
    {
      printf ("\n  unable to calibrate, loop limit out of range\n");
      exit (EXIT_FAILURE);
    }
    goto loop;
  }
  arguments [ARG_N_USLN].value = (void *) (n / i);
  printf (" n = %lu clks\n", ARG_USLN);
  usleep_cal = usleep_loop;
}

ARGUMENT_CALLBACK (usleep_setn)
{
  long n;
  char *tail;
  n = strtol (parameter, &tail, 0);
  if (*tail != 0)
    arguments_error (arg, "requires a number as parameter");
  if (n < 0)
    arguments_error (arg, "requires a positive value");
  arguments [ARG_N_USLN].value = (void *) n;
  usleep_cal = usleep_loop;
}

ARGUMENT_CALLBACK (argument_clkt)
{
  long n;
  char *tail;
  n = strtol (parameter, &tail, 0);
  if ((*tail != 0) || (n < 2) || (n > 1000))
    arguments_error (arg, "requires a number within the range of 2-1000.");
  arguments [ARG_N_CLKT].value = (void *) (n >> 1);
}



// Status bar/line for showing operation progress

static void status_bar_flow (unsigned at, unsigned p, unsigned length)
{
  if (p == length)
    printf ("%u\n", length);
  else if (p % 128 == 0)
    printf ("%u", p);
  else if (p % 16 == 0)
    putchar ('.');
}

time_t stat_last_t = 0;
unsigned stat_bits, stat_in, stat_out;

#define stat_reset() \
  { \
    stat_bits = 0; \
    stat_in = 0; \
    stat_out = 0; \
  }


static void status_bar_progress (unsigned at, unsigned p, unsigned length)
{
  if (p == length)
    printf ("\r    %u  %u/%u  100%%\e[K\n", at, p, length);
  else
  {
    time_t t;
    t = time (NULL);
    if (t != stat_last_t)
    {
      stat_last_t = t;
      printf ("\r    %u  %u/%u  %u%%   %u bps  %u/%u B/s\e[K", at, p, length, 
         p * 100 / length, stat_bits, stat_in, stat_out);
      stat_reset ();
    }
  }
}


ARGUMENT_CALLBACK (status_progress)
{
  arguments [ARG_N_STAT].value = status_bar_progress;
}

ARGUMENT_CALLBACK (status_flow)
{
  arguments [ARG_N_STAT].value = status_bar_flow;
}




// config bits

static void show_bits (const NameID *nid)
{
  while (nid -> id != 0)
  {
    printf ("  0x%02X %s (%s)\n", nid -> id, nid -> name, nid -> short_name);
    nid ++;
  }
}

static void show_bits_set (const NameID *nid, unsigned char bits, char *nobits)
{
  unsigned i;
  
  if (bits == 0xff)
    printf ("  (%s)\n", nobits);
  else
  {
    for (i = 0; nid [i].id != 0; i ++)
      if ((bits & nid [i].id) == 0)
      {
        printf ("  [%u] %s\n", i, nid [i].name);
        bits |= nid [i].id;
      }
    for (i = 0; bits != 0xff; bits = (bits >> 1) | 0x80)
      if ((bits & 1) == 0)
        i ++;
    if (i > 0)
      printf ("  + %u unknown\n", i);
  }
}


static int set_bits (const NameID *nid, const char *arg, unsigned char *bits)
{
  unsigned l, i;
  char *tail;
  
  *bits = 0xff;
loop:
  for (l = 0; (arg [l] != 0) && (arg [l] != ','); l ++);
  if (l > 0)
  {
    i = strtol (arg, &tail, 0);
    if (tail == arg + l)
      *bits &= i;
    else
    {
      i = 0;
      while (strncasecmp (arg, nid [i].short_name, l) != 0)
      {
        if (nid [++ i].id == 0)
        {
          printf ("Invalid lock bit name \"");
          while (l -- > 0)
            putchar (*arg ++);
          puts ("\"");
          return -1;
        }
      }
      *bits &= ~nid [i].id;
    }
  }
  if (arg [l] == ',')
  {
    arg += l + 1;
    goto loop;
  }
  return 0;
}

